/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.yuandian.dataflow.statemachine;

import org.apache.http.entity.InputStreamEntity;
import org.apache.ratis.proto.RaftProtos;
import org.apache.ratis.proto.RaftProtos.RaftConfigurationProto;
import org.apache.ratis.proto.RaftProtos.RoleInfoProto;
import org.apache.ratis.protocol.Message;
import org.apache.ratis.protocol.RaftGroupId;
import org.apache.ratis.protocol.RaftGroupMemberId;
import org.apache.ratis.protocol.RaftPeer;
import org.apache.ratis.protocol.RaftPeerId;
import org.apache.ratis.server.RaftServer;
import org.apache.ratis.server.protocol.TermIndex;
import org.apache.ratis.server.raftlog.RaftLog;
import org.apache.ratis.server.storage.RaftStorage;
import org.apache.ratis.statemachine.TransactionContext;
import org.apache.ratis.statemachine.impl.BaseStateMachine;
import org.apache.ratis.statemachine.impl.SimpleStateMachineStorage;
import org.apache.ratis.statemachine.impl.SingleFileSnapshotInfo;
import org.apache.ratis.util.AutoCloseableLock;
import org.apache.ratis.util.JavaUtils;

import com.yuandian.dataflow.statemachine.master.MasterFactory;
import com.yuandian.dataflow.statemachine.messages.MessageUtils;
import com.yuandian.dataflow.statemachine.messages.Operate;
import com.yuandian.dataflow.statemachine.messages.Query;
import com.yuandian.dataflow.statemachine.messages.RaftReply;
import com.yuandian.dataflow.statemachine.messages.Operate.OperateType;
import com.yuandian.dataflow.statemachine.messages.RaftReply.Status;
import com.yuandian.dataflow.statemachine.state.State;
import com.yuandian.dataflow.statemachine.state.WorkerState;

import lombok.extern.slf4j.Slf4j;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.nio.charset.Charset;
import java.util.Collection;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantReadWriteLock;

/**
 * State machine implementation for Counter server application. This class
 * maintain a {@link AtomicInteger} object as a state and accept two commands:
 * GET and INCREMENT, GET is a ReadOnly command which will be handled by
 * {@code query} method however INCREMENT is a transactional command which
 * will be handled by {@code applyTransaction}.
 */
@Slf4j
public class StateMachine extends BaseStateMachine {
  private final SimpleStateMachineStorage storage = new SimpleStateMachineStorage();

  private State state = new State();

  private AtomicInteger counter = new AtomicInteger(0);

  private AtomicBoolean leader = new AtomicBoolean(false);

  private final ReentrantReadWriteLock lock = new ReentrantReadWriteLock(true);

  private AutoCloseableLock readLock() {
    return AutoCloseableLock.acquire(lock.readLock());
  }

  private AutoCloseableLock writeLock() {
    return AutoCloseableLock.acquire(lock.writeLock());
  }

  public Boolean isLeader() {
    return leader.get();
  }

  public Executor asyncExecutor = Executors.newFixedThreadPool(8);

  /**
   * initialize the state machine by initilize the state machine storage and
   * calling the load method which reads the last applied command and restore it
   * in counter object)
   *
   * @param server      the current server information
   * @param groupId     the cluster groupId
   * @param raftStorage the raft storage which is used to keep raft related
   *                    stuff
   * @throws IOException if any error happens during load state
   */
  @Override
  public void initialize(RaftServer server, RaftGroupId groupId,
      RaftStorage raftStorage) throws IOException {
    super.initialize(server, groupId, raftStorage);
    this.storage.init(raftStorage);
    load(storage.getLatestSnapshot());
  }

  /**
   * very similar to initialize method, but doesn't initialize the storage
   * system because the state machine reinitialized from the PAUSE state and
   * storage system initialized before.
   *
   * @throws IOException if any error happens during load state
   */
  @Override
  public void reinitialize() throws IOException {
    load(storage.getLatestSnapshot());
  }

  /**
   * Store the current state as an snapshot file in the stateMachineStorage.
   *
   * @return the index of the snapshot
   */
  @Override
  public long takeSnapshot() {
    // get the last applied index
    final TermIndex last = getLastAppliedTermIndex();

    // create a file with a proper name to store the snapshot
    final File snapshotFile = storage.getSnapshotFile(last.getTerm(), last.getIndex());

    // serialize the counter object and write it into the snapshot file
    try {
      ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(snapshotFile)));
      out.writeObject(counter);
      out.close();
    } catch (IOException ioe) {
      LOG.warn("Failed to write snapshot file \"" + snapshotFile
          + "\", last applied index=" + last);
    }

    // return the index of the stored snapshot (which is the last applied one)
    return last.getIndex();
  }

  /**
   * Load the state of the state machine from the storage.
   *
   * @param snapshot to load
   * @return the index of the snapshot or -1 if snapshot is invalid
   * @throws IOException if any error happens during read from storage
   */
  private long load(SingleFileSnapshotInfo snapshot) throws IOException {
    // check the snapshot nullity
    if (snapshot == null) {
      LOG.warn("The snapshot info is null.");
      return RaftLog.INVALID_LOG_INDEX;
    }

    // check the existance of the snapshot file
    final File snapshotFile = snapshot.getFile().getPath().toFile();
    if (!snapshotFile.exists()) {
      LOG.warn("The snapshot file {} does not exist for snapshot {}",
          snapshotFile, snapshot);
      return RaftLog.INVALID_LOG_INDEX;
    }

    // load the TermIndex object for the snapshot using the file name pattern of
    // the snapshot
    final TermIndex last = SimpleStateMachineStorage.getTermIndexFromSnapshotFile(snapshotFile);

    // read the file and cast it to the AtomicInteger and set the counter
    try (ObjectInputStream in = new ObjectInputStream(
        new BufferedInputStream(new FileInputStream(snapshotFile)))) {
      // set the last applied termIndex to the termIndex of the snapshot
      setLastAppliedTermIndex(last);

      // read, cast and set the counter
      counter = JavaUtils.cast(in.readObject());
    } catch (ClassNotFoundException e) {
      throw new IllegalStateException(e);
    }

    return last.getIndex();
  }

  @Override
  public void notifyLeaderChanged(RaftGroupMemberId groupMemberId, RaftPeerId newLeaderId) {

    log.info("msg {}", groupMemberId.getPeerId());
    if (StateServerFactory.getCurrentPeerId().getRaftPeerId() == null) {
      StateServerFactory.getCurrentPeerId().setRaftPeerId(groupMemberId.getPeerId().toString());
    }

    leader.set(newLeaderId == groupMemberId.getPeerId());

    log.info("newLeaderId: {} groupMemberId: {} leader {}", newLeaderId, groupMemberId.getPeerId(), isLeader());

    asyncExecutor.execute(() -> {
      log.info("asyncExecutor");
      var op = new Operate(OperateType.PUT_WORKERSTATE, new WorkerState(StateServerFactory.getCurrentPeerId()));
      try {
        var reply = StateServerFactory.send(op);
        log.info("{}",  MessageUtils.<RaftReply>fromMessage(reply.getMessage()));
      } catch (IOException | ClassNotFoundException e) {
        log.error("{}",e.toString());
      }
    });

    if (MasterFactory.getMasterExecute().isAlive())
      MasterFactory.getMasterExecute().interrupt();

    if (isLeader())
      MasterFactory.getMasterExecute().start();

  }

  /**
   * Handle GET command, which used by clients to get the counter value.
   *
   * @param request the GET message
   * @return the Message containing the current counter value
   */
  @Override
  public CompletableFuture<Message> query(Message request) {

    var reply = new RaftReply();
 
    try {
      var query = MessageUtils.<Query>fromMessage(request);
      switch (query.getType()) {
        case GET_STATE:
          try (var rlock = readLock()) {
            var ws = state.getWorkers().get(((WorkerState) query.getValue()).getPeerId());
            if (ws == null) {
              return CompletableFuture.completedFuture(Status.setError(reply, "Peerid is not exist"));
            }
            reply.setData(ws); 
          }
          break;
        default:
    
          Status.setError(reply, "Invalid Command");
          return CompletableFuture.completedFuture(reply);

      }

    } catch (ClassNotFoundException | IOException e) {
      log.error("{}", e.toString());
    }
 
    return CompletableFuture.completedFuture(Status.setOK(reply));
  }

  /**
   * Apply the INCREMENT command by incrementing the counter object.
   *
   * @param trx the transaction context
   * @return the message containing the updated counter value
   */
  @Override
  public CompletableFuture<Message> applyTransaction(TransactionContext trx) {
    // log.info("applyTransaction");
    var reply = new RaftReply();

    final RaftProtos.LogEntryProto entry = trx.getLogEntry();
    final long index = entry.getIndex();
  
    try {

      var data = entry.getStateMachineLogEntry().getLogData();
      final var op = MessageUtils.<Operate>fromByteString(data);
      try (var wlock = writeLock()) {
  
        switch (op.getType()) {
          case ALLOCATE_PACKETS:
            break;
          case PUT_WORKERSTATE:
            var ws = (WorkerState) op.getValue();
            if(ws.getPeerId().getRaftPeerId() == null) {
              return CompletableFuture.completedFuture(Status.setError(reply, "RaftPeerId 为 null"));
            }
            state.getWorkers().put(ws.getPeerId(), ws);
            break;
          case REMOVE:
            break;
          default:
            break;
        }
        // if (isLeader()) {
        //   log.info("{}: getType   {} ", index, op.getType());
        // }
        log.info("{}: getType   {} ", index, op.getType());
        updateLastAppliedTermIndex(entry.getTerm(), index);
      }
 
    } catch (IOException | ClassNotFoundException e) {
      log.info("{}", e.toString());
      return CompletableFuture.completedFuture(Status.setError(reply, "错误op"));
    }
 
    return CompletableFuture.completedFuture(Status.setOK(reply));
  }
}
